############################################################################
#                                                                          #
# by jackal | me@jackal.in | 2007-05-09 | v0.1                             #
#                                                                          #
#   .                                         #
#                                                                          #
############################################################################
package jackal::Parser::Referrer;
use strict;
use warnings;
use vars qw(@ISA @EXPORT @EXPORT_OK);

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(analyze);
@EXPORT_OK = ();

my @cfg_se = qw(yandex rambler google mail msn aport yahoo webalta nigma altavista);
my @cfg_params = qw(text words query q r p st request_str qs);
my @cfg_noparams = qw(btng client old_q hid btnf ses_id sess session phpsessid);

#---------------------------
use URI::Escape;
use Encode qw(decode);
use Lingua::DetectCharset;
use Convert::Cyrillic;
$Lingua::DetectCharset::DoubtLog = undef;

#  ,     
#     QP (    )
#   cp1251
sub analyze {
    my $ref = shift;
    my ($host, $param) = $ref =~ m#^https?://([^\?/]+)/(?:[^\?]*)\?(.+)$#i;
    return undef unless $host;
    $host =~ s#^www\.##;
    my $se = detect_se($host);
    return ($se, undef) unless $param;
    my @params;
    while ($param =~ /([^\=\&\;]+)=([^\=\&\;]+)/g) {
        push(@params, {name => lc($1), param => unescape($2)});
    }
    $param = detect_param(\@params);

    return ($se, $param);
}

#################################

#    
sub unescape {
    my $s = shift;
    $s =~ s/\+/ /sg;
    $s = uri_unescape($s);
    $s =~ s/^\s+//;
    $s =~ s/\s+$//;
    $s =~ s/\s{2,}/ /g;
    return $s;
}

#  
sub detect_se {
    my $host = shift;
    $host = lc($host);
    foreach my $s (@cfg_se) {
        return $s if $host =~ /$s(?:\.[a-z]{2,4}){1,2}$/;
    }
    return 'other';
}

#      
sub detect_param {
    my $params = shift;
    my @pg;
    my @pb;
    foreach my $p (@$params) {
        my $ok = 0;
        foreach my $c (@cfg_params) {
            if ($p->{name} eq $c) {
                push(@pg, $p->{param});
                $ok = 1;
                last;
            }
        }
        next if $ok;

        $ok = 1;
        foreach my $c (@cfg_noparams) {
            if ($p->{name} eq $c) {
                $ok = 0;
                last;
            }
        }
        push (@pb, $p->{param}) if $ok;
    }

    my $r = undef;
    if ($#pg == 0) {
        $r = analyze_param($pg[0]);
    } elsif ($#pg > 0) {
        $r = heuristic_param(\@pg);
    }
    return $r if $r;

    if ($#pb == 0) {
        return analyze_param($pb[0]);
    } elsif ($#pb > 0) {
        return heuristic_param(\@pb);
    } else {
        return undef;
    }
}

#      
sub heuristic_param {
    my $arref = shift;
    my @res;
    foreach my $p (@$arref) {
        my $r = analyze_param($p);
        next unless $r;
        push(@res, $r);
    }
    local *h_length = sub {
        my $s = shift;
        my $c = 0;
        $c++ while ($s =~ /[--]/g);
        return length($s)+$c*8;
    };
    if ($#res == 0) {
        return $res[0];
    } elsif ($#res > 0) {
        my @arr = sort {h_length($b) <=> h_length($a)} @res;
        return $arr[0];
    } else {
        return undef;
    }
}

#   - , QP   
sub analyze_param {
    my $param = shift;
    return undef if $param =~ /^\d+$/;
    return undef if lc($param) =~ /^[a-f0-9]{8,}$/;

    #  QP-  
    my @q_params;
    while ($param =~ /([^\=\&\;]+)=([^\=\&\;]+)/g) {
        push(@q_params, {name => lc($1), param => unescape($2)});
    }
    if ($#q_params == 0) {
        $param = $q_params[0];
    } elsif ($#q_params > 0) {
        return detect_param(\@q_params);
    }

    #  QP
    my $c = 0;
    $c++ while ($param =~ /\%[A-Fa-f0-9]{2}/g);
    $param = unescape($param) if ($c > 1);

    #  UTF-16BE
    local *utf16be = sub {
        my $s = shift;
        $s =~ s/\%u([A-Fa-f0-9]{4})/pack("H4", $1)/ge;
        $s = decode("UTF-16BE", $s);
        $s = Convert::Cyrillic::cstocs("UTF8", "WIN", $s);
        return $s;
    };
    goto DONE if $param =~ s/((?:\%u[A-Fa-f0-9]{4})+)/utf16be($1)/ge;

    #  
    my $charset = Lingua::DetectCharset::Detect($param);
    $charset = "UTF8" if $charset eq "ENG";
    if ($charset eq "KOI8") { #     - workaround
        my $p = Convert::Cyrillic::cstocs("WIN", "KOI8", $param);
        $charset = Lingua::DetectCharset::Detect($p);
        $charset = ($charset ne "WIN") ? "WIN" : "KOI8";
    }
    if ($charset ne "WIN"){
        $param = Convert::Cyrillic::cstocs($charset, "WIN", $param);
    }

    DONE:
    return undef if length($param) < 3;
    return $param;
}

1;

